#pragma once

#include "invokee.hpp"
#include "quake_camera.hpp"
#include "timer_interface.hpp"
#include <vector>
#include <glm/gtx/norm.hpp> // For glm::distance2 (squared distance)

class camera_path : public avk::invokee
{
public:
  camera_path(avk::quake_camera &cam)
      : mCam{ &cam } // Target camera
        ,
        mSpeed{ 2.0f } // Movement speed
        ,
        mStartTime{} // Set in initialize()
        ,
        mCurrentWaypointIndex(0)
  {
    // Initial waypoints with positions to move to and look-at targets
    mWaypoints.push_back(
        { glm::vec3(2.0f, 2.0f, 33.0f), glm::vec3(2.0f, 2.0f, 24.0f) });
    //mWaypoints.push_back(
    //    { glm::vec3(2.0f, 2.0f, 32.0f), glm::vec3(2.0f, 2.0f, 24.0f) });
    mWaypoints.push_back(
        { glm::vec3(2.0f, 2.0f, 32.0f), glm::vec3(-0.5f, 2.0f, 24.0f) });   //left painting 2-set
    mWaypoints.push_back(
        { glm::vec3(2.0f, 2.0f, 29.0f), glm::vec3(3.5f, 2.0f, 24.0f) });    //right painting 2-set
    mWaypoints.push_back(
        { glm::vec3(10.0f, 2.0f, 27.0f), glm::vec3(16.0f, 2.0f, 27.5f) });  //left painting 2nd 2-set
    mWaypoints.push_back(
        { glm::vec3(10.0f, 2.0f, 25.0f), glm::vec3(14.0f, 2.0f, 19.0f) });  //right painting 2nd 2-set
    mWaypoints.push_back(
        { glm::vec3(12.0f, 2.0f, 23.0f), glm::vec3(18.0f, 2.0f, 19.0f) });
    mWaypoints.push_back(
        { glm::vec3(7.0f, 2.0f, 20.0f), glm::vec3(1.5f, 2.0f, 16.5f) });    //last painting single
    mWaypoints.push_back(
        { glm::vec3(10.0f, 2.0f, 14.0f), glm::vec3(10.0f, 2.0f, 9.0f) });
    mWaypoints.push_back(
        { glm::vec3(10.0f, 2.0f, 10.0f), glm::vec3(16.0f, 2.0f, 2.0f) });
    mWaypoints.push_back(
        { glm::vec3(15.0f, 2.0f, 8.0f), glm::vec3(18.0f, 2.5f, 8.0f) });    //look at bunny
    mWaypoints.push_back(
        { glm::vec3(16.0f, 2.0f, 0.0f), glm::vec3(19.0f, 3.0f, -8.0f) });
    mWaypoints.push_back(
        { glm::vec3(10.0f, 2.0f, -6.0f), glm::vec3(1.0f, 5.0f, 0.0f) });    //look around dragon
    mWaypoints.push_back(
        { glm::vec3(5.0f, 2.0f, -6.0f), glm::vec3(0.0f, 5.0f, 0.0f) });
    mWaypoints.push_back(
        { glm::vec3(0.0f, 2.0f, -10.0f), glm::vec3(0.0f, 4.0f, 0.0f) });    //back of dragon
    mWaypoints.push_back(
        { glm::vec3(-5.0f, 2.0f, -6.0f), glm::vec3(-5.0f, 4.0f, -0.5f) });   // transition away dragon
    mWaypoints.push_back(
        { glm::vec3(-10.0f, 2.0f, -6.0f), glm::vec3(-18.0f, 2.0f, 0.5f) });
    mWaypoints.push_back(
        { glm::vec3(-14.0f, 2.0f, 9.0f), glm::vec3(-14.0f, 2.0f, 11.0f) });
    mWaypoints.push_back(
        { glm::vec3(-14.0f, 2.0f, 14.0f), glm::vec3(-14.0f, -2.0f, 20.0f) });
    mWaypoints.push_back(
        { glm::vec3(-14.0f, 2.0f, 14.0f), glm::vec3(-14.0f, -2.0f, 20.0f) });    //dummy end at last position
  }

  int execution_order() const override
  {
    return 5; // Run after quake_camera to prevent jittering from mouse
              // movements
  }


  // Function to calculate the total distance of the entire path
  void calculateTotalDistance()
  {
    mTotalDistance = 0.0f;
    for (size_t i = 0; i < mWaypoints.size() - 1; ++i)
    {
      mTotalDistance
          += glm::distance(mWaypoints[i].first, mWaypoints[i + 1].first);
    }
  }


  void initialize() override
  {
    mStartTime = avk::time().time_since_start();

    // Set the camera's initial position to the first waypoint
    if (!mWaypoints.empty())
    {
      mCam->set_translation(mWaypoints[0].first);
      mCam->look_at(mWaypoints[0].second);
      mCurrentLookAt = mWaypoints[0].second; // Set initial look-at
      calculateTotalDistance();
    }
  }

  // Function to interpolate the look-at direction using spherical
  // interpolation (slerp)
  glm::vec3 interpolateLookAt(const glm::vec3 &currentLookAt,
      const glm::vec3 &targetLookAt, glm::vec3 currentPos, glm::vec3 targetPos, float t)
  {
    // Normalize the current and target look-at directions to ensure they are
    // unit vectors
    glm::vec3 normalizedCurrentDirection
        = glm::normalize(currentLookAt - currentPos);
    glm::vec3 normalizedTargetDirection
        = glm::normalize(targetLookAt - targetPos);


    glm::vec3 interpolatedLookAtDirection = glm::normalize(
        glm::mix(normalizedCurrentDirection, normalizedTargetDirection, t));

    return interpolatedLookAtDirection;
  }

  void update() override
  {
    float elapsedTime = avk::time().time_since_start() - mStartTime;

    if (mCurrentWaypointIndex >= mWaypoints.size()-1)
      return;

    // Retrieve the current target position and target look-at position

    glm::vec3 originPosition = mWaypoints[mCurrentWaypointIndex].first;
    glm::vec3 originLookAt = mWaypoints[mCurrentWaypointIndex].second;

    glm::vec3 targetPosition = mWaypoints[mCurrentWaypointIndex+1].first;
    glm::vec3 targetLookAt = mWaypoints[mCurrentWaypointIndex+1].second;

    glm::vec3 currentPosition = mCam->translation();

    // Direction for this segment
    glm::vec3 direction = glm::normalize(targetPosition - currentPosition);


    auto deltaSpeed = mSpeed * avk::time().delta_time();
    // move into this direction
    glm::vec3 newPosition = currentPosition + direction * deltaSpeed;

    //distance of segment and the one so far travelled on that segment
    float segmentDistance = glm::distance(originPosition, targetPosition);
    float segmentDistanceTraveled = glm::distance(originPosition, newPosition);

    //t = [0,1] calculate how far along the segment has been travelled
    float interpolationFactor
        = glm::clamp(segmentDistanceTraveled / segmentDistance, 0.0f, 1.0f);

    //calculate direction to look at in this step. based on directions origin and target
    //direction calculated via lookAt-targets and positions
    glm::vec3 interpolatedLookAtDirection = interpolateLookAt(originLookAt,
        targetLookAt, originPosition, targetPosition, interpolationFactor);

    // set the new lookAt-target
    mCam->look_along(interpolatedLookAtDirection);

    // set the new position of the camera
    mCam->set_translation(newPosition);

    //check if segment goal has been reached
    bool positionReached = glm::distance2(newPosition, targetPosition)
                           < deltaSpeed * deltaSpeed; // larger than movespeed
   
    // if segment goal has been reached, stop and set the next one
    if (positionReached)
    {
      newPosition = targetPosition; // Set position exactly to target
      mCurrentWaypointIndex++;      // Move to next waypoint
    }    
  }


  // Method to add a waypoint with both position and look-at target
  void set_waypoint(const glm::vec3 &position, const glm::vec3 &lookAt)
  {
    mWaypoints.push_back({ position, lookAt });
  }

  // Method to set the camera movement speed
  void set_speed(float speed)
  {
    mSpeed = speed;
  }


private:
  avk::quake_camera *mCam;      // Reference to the camera
  float mSpeed;                 // Movement speed
  float mStartTime;             // Start time for movement
  size_t mCurrentWaypointIndex; // Index of the current waypoint
  std::vector<std::pair<glm::vec3, glm::vec3>>
      mWaypoints; // List of waypoints with position and look-at target

  glm::vec3 mCurrentLookAt; // Current look-at target direction for smooth
                            // interpolation

  float mTotalDistance;     // total over all waypoints
  float mDistanceTraveled; // Distance traveled so far
};
