#include "imgui.h"

#include "quake_camera.hpp"
#include "configure_and_compose.hpp"
#include "invokee.hpp"
#include "material_image_helpers.hpp"
#include "model.hpp"
#include "sequential_invoker.hpp"
#include "vk_convenience_functions.hpp"
#include <AL/al.h>
#include <AL/alc.h>
#include <sndfile.h>
#include "camera_path.hpp"

#include "scene.h"
#include "shadows.h"

constexpr std::array CLEAR_COLOR = { 0.53f, 0.81f, 0.92f, 1.f };

class Demo : public avk::invokee
{
  struct ModelMatrices
  {
    glm::mat4 mModelMatrix;
    glm::mat4 mNormalMatrix;
  };

  struct alignas(16) PushConstantsLightPass
  {
    int debugMode;                        // 4 bytes
    alignas(16) glm::vec3 cameraPosition; // 16 bytes
  };

#define NUM_LIGHTS 24

  // alignas probably useless now since vec4 is used, but in case we change
  // that back to vec3 everything as vec4 since arrays get padded with std140
  // to 16bytes which is 4*4bytes for a vec4
  struct LightData
  {
    glm::vec4 lightPositions[NUM_LIGHTS];  // Position of lights
    glm::vec4 lightColors[NUM_LIGHTS];     // Color of lights
    glm::vec4 lightDirections[NUM_LIGHTS]; // Directions for spotlights
    glm::vec4 radius[NUM_LIGHTS];          // Attenuation radius
    glm::vec4 coneAngles[NUM_LIGHTS];      // Cone angles for spotlights
  };

  struct VolumetricLightData
  {
    glm::vec4 lightPositions;
    glm::vec4 cameraPosition;
    glm::vec4 params;
  };

  // Audio info for background music
  struct AudioData
  {
    std::vector<short> samples;
    int sampleRate;
    int channels;
  };

public
    : // v== avk::invokee overrides which will be invoked by the framework ==v
  Demo(avk::queue &aQueue) : mQueue{ &aQueue }, mScale{ 10.0f, 10.0f, 10.0f }
  {
  }

  void initializeOpenAL(ALCdevice *&device, ALCcontext *&context)
  {
    device = alcOpenDevice(nullptr); // Default device
    if (!device)
    {
      throw std::runtime_error("Failed to open OpenAL device.");
    }

    context = alcCreateContext(device, nullptr);
    if (!alcMakeContextCurrent(context))
    {
      throw std::runtime_error("Failed to set OpenAL context.");
    }
  }

  AudioData loadAudioFile(const std::string &filePath)
  {
    SF_INFO fileInfo{};
    SNDFILE *file = sf_open(filePath.c_str(), SFM_READ, &fileInfo);
    if (!file)
    {
      throw std::runtime_error("Failed to open audio file: " + filePath);
    }

    std::vector<short> samples(fileInfo.frames * fileInfo.channels);
    sf_read_short(file, samples.data(), samples.size());
    sf_close(file);

    return { samples, fileInfo.samplerate, fileInfo.channels };
  }

  void playAudio(const AudioData &audioData)
  {
    ALuint buffer, source;
    alGenBuffers(1, &buffer);
    alGenSources(1, &source);

    ALenum format
        = (audioData.channels == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
    alBufferData(buffer, format, audioData.samples.data(),
        audioData.samples.size() * sizeof(short), audioData.sampleRate);

    alSourcei(source, AL_BUFFER, buffer);
    alSourcei(source, AL_LOOPING, AL_TRUE); // Enable looping

    alSourcePlay(source);

    ALint state;
    do
    {
      alGetSourcei(source, AL_SOURCE_STATE, &state);
    } while (state == AL_PLAYING);

    // Clean up
    alDeleteSources(1, &source);
    alDeleteBuffers(1, &buffer);
  }

  void startAudioPlayback(Demo &app, const AudioData &audioData)
  {
    std::thread audioThread(
        [&app, &audioData]() { app.playAudio(audioData); });
    audioThread.detach(); // Detach the thread to run independently
  }

  void cleanupOpenAL(ALCdevice *device, ALCcontext *context)
  {
    alcMakeContextCurrent(nullptr);
    alcDestroyContext(context);
    alcCloseDevice(device);
  }

  void initialize() override
  {

    // start camera path
    mCameraPath.emplace(mQuakeCam);
    avk::current_composition()->add_element(mCameraPath.value());

    mInitTime = std::chrono::high_resolution_clock::now();

    mDescriptorCache = avk::context().create_descriptor_cache();

    mScene = Scene::load("assets/museum.gltf", mQueue);

    mViewProjBuffer
        = avk::context().create_buffer(avk::memory_usage::host_coherent, {},
            avk::uniform_buffer_meta::create_from_data(glm::mat4()));

    // Temporary, will be replaced if needed
    auto const modelMatrix = glm::mat4(1.0);
    auto const normalMatrix = glm::transpose(glm::inverse(modelMatrix));
    auto const modelMatrices = ModelMatrices{ modelMatrix, normalMatrix };
    mModelMatrices
        = avk::context().create_buffer(avk::memory_usage::host_coherent, {},
            avk::uniform_buffer_meta::create_from_data(ModelMatrices{}));
    // its on the host, no need to sync
    auto _ignore = mModelMatrices->fill(&modelMatrices, 0);

    const auto res = avk::context().main_window()->resolution();

    mShadows = Shadows::create(
        mlightData.lightPositions, mlightData.lightDirections, res, mQueue);
    mShadows.render(mScene, mQueue);

    /*********************************************
    **********************************************
    DEFERRED SHADING
    **********************************************
    *********************************************/

    // light setup
    mLightBuffer = avk::context().create_buffer(avk::memory_usage::device, {},
        avk::uniform_buffer_meta::create_from_size(sizeof(mlightData)));

    auto lightCmd
        = mLightBuffer->fill(reinterpret_cast<const void *>(&mlightData), 0);

    auto lightFence = avk::context().record_and_submit_with_fence(
        { std::move(lightCmd) }, *mQueue);
    lightFence->wait_until_signalled();

    // for deffered shading
    // Create G-Buffer images for positions, normals, albedo, depth

    auto gBufferPosition = avk::context().create_image(res.x, res.y,
        { vk::Format::eR32G32B32A32Sfloat, vk::SampleCountFlagBits::e1 }, 1,
        avk::memory_usage::device, avk::image_usage::general_color_attachment);

    auto gBufferAlbedo = avk::context().create_image(res.x, res.y,
        { vk::Format::eR8G8B8A8Unorm, vk::SampleCountFlagBits::e1 }, 1,
        avk::memory_usage::device, avk::image_usage::general_color_attachment);

    auto gBufferNormal = avk::context().create_image(res.x, res.y,
        { vk::Format::eR16G16B16A16Sfloat, vk::SampleCountFlagBits::e1 }, 1,
        avk::memory_usage::device, avk::image_usage::general_color_attachment);

    auto gBufferDepth = avk::context().create_depth_stencil_image(res.x, res.y,
        vk::Format::eD32Sfloat, 1, avk::memory_usage::device,
        avk::image_usage::general_depth_stencil_attachment);

    // Create image_views and image_samplers for each image
    // put image_samplers into image_sampler vector

    auto sampler = avk::context().create_sampler(
        avk::filter_mode::trilinear, avk::border_handling_mode::repeat);

    mGBufferPositionView
        = avk::context().create_image_view(gBufferPosition, {});
    mGBufferImageSamplers.push_back(
        avk::context().create_image_sampler(mGBufferPositionView, sampler));

    mGBufferAlbedoView = avk::context().create_image_view(gBufferAlbedo, {});
    mGBufferImageSamplers.push_back(
        avk::context().create_image_sampler(mGBufferAlbedoView, sampler));

    mGBufferNormalView = avk::context().create_image_view(gBufferNormal, {});
    mGBufferImageSamplers.push_back(
        avk::context().create_image_sampler(mGBufferNormalView, sampler));

    mGBufferDepthView = avk::context().create_image_view(gBufferDepth, {});
    mGBufferImageSamplers.push_back(
        avk::context().create_image_sampler(mGBufferDepthView, sampler));

    // Create attachments for each and framebuffer with that image views
    // automatically creates renderpass inside create_framebuffer()-call

    auto gBufferPositionAttachment
        = avk::attachment::declare_for(mGBufferPositionView.as_reference(),
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::color(0), avk::on_store::store);

    auto gBufferAlbedoAttachment
        = avk::attachment::declare_for(mGBufferAlbedoView.as_reference(),
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::color(1), avk::on_store::store);

    auto gBufferNormalAttachment
        = avk::attachment::declare_for(mGBufferNormalView.as_reference(),
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::color(2), avk::on_store::store);

    auto gBufferDepthAttachment
        = avk::attachment::declare_for(mGBufferDepthView.as_reference(),
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::depth_stencil, avk::on_store::store);

    // Create framebuffer from attachments  and views (need to move into
    // vectors)

    mGBufferFramebuffer = avk::context().create_framebuffer(
        { gBufferPositionAttachment, gBufferAlbedoAttachment,
            gBufferNormalAttachment, gBufferDepthAttachment },
        { mGBufferPositionView, mGBufferAlbedoView, mGBufferNormalView,
            mGBufferDepthView });

    // might be able to use the one used for deferred
    auto volumetricSampler = avk::context().create_sampler(
        avk::filter_mode::trilinear, avk::border_handling_mode::repeat);

    // data for buffer
    mVolumetricLightBuffer
        = avk::context().create_buffer(avk::memory_usage::device, {},
            avk::uniform_buffer_meta::create_from_size(
                sizeof(mVolumetricLightData)));

    auto volumetricLightCmd = mVolumetricLightBuffer->fill(
        reinterpret_cast<const void *>(&mVolumetricLightData), 0);

    auto volumetricLightFence = avk::context().record_and_submit_with_fence(
        { std::move(volumetricLightCmd) }, *mQueue);
    volumetricLightFence->wait_until_signalled();

    mInvViewProjBuffer
        = avk::context().create_buffer(avk::memory_usage::host_coherent, {},
            avk::uniform_buffer_meta::create_from_data(glm::mat4()));

    // texture preparation
    auto volumetricLightImage = avk::context().create_image(res.x, res.y,
        { vk::Format::eR16G16B16A16Sfloat, vk::SampleCountFlagBits::e1 }, 1,
        avk::memory_usage::device, avk::image_usage::general_color_attachment);

    mVolumetricLightView
        = avk::context().create_image_view(volumetricLightImage, {});

    mVolumetricLightingImageSamplers.push_back(
        avk::context().create_image_sampler(
            mVolumetricLightView, volumetricSampler));

    auto volumetricLightAttachment
        = avk::attachment::declare_for(mVolumetricLightView.as_reference(),
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::color(0), avk::on_store::store);

    // framebuffer
    mVolumetricLightFramebuffer = avk::context().create_framebuffer(
        { volumetricLightAttachment }, { mVolumetricLightView });

    /*********************************************
    GEOMETRY
    *********************************************/

    //  Create our rasterization graphics pipeline with the required
    //  configuration:
    mGBufferPipeline = avk::context().create_graphics_pipeline_for(
        // Specify which shaders the pipeline consists of:
        avk::vertex_shader("shaders/def_geometry.vert"),
        avk::fragment_shader("shaders/def_geometry.frag"),
        // The next 3 lines define the format and location of the vertex shader
        // inputs: (The dummy values (like glm::vec3) tell the pipeline the
        // format of the respective input)
        avk::from_buffer_binding(0)
            ->stream_per_vertex<glm::vec3>()
            ->to_location(0), // <-- corresponds to vertex shader's inPosition
        avk::from_buffer_binding(1)
            ->stream_per_vertex<glm::vec2>()
            ->to_location(1), // <-- corresponds to vertex shader's inTexCoord
        avk::from_buffer_binding(2)
            ->stream_per_vertex<glm::vec3>()
            ->to_location(2), // <-- corresponds to vertex shader's inNormal
        avk::from_buffer_binding(3)
            ->stream_per_vertex<glm::vec3>()
            ->to_location(3), // <-- corresponds to vertex shader's inNormal
        avk::from_buffer_binding(4)
            ->stream_per_vertex<glm::vec3>()
            ->to_location(4), // <-- corresponds to vertex shader's inNormal

        // Some further settings:
        avk::cfg::front_face::define_front_faces_to_be_counter_clockwise(),
        avk::cfg::viewport_depth_scissors_config::from_framebuffer(
            mGBufferFramebuffer.as_reference()),
        gBufferPositionAttachment, gBufferAlbedoAttachment,
        gBufferNormalAttachment, gBufferDepthAttachment,
        // Additional data passed to shaders (constant to vertex
        // shader_type::vertex,
        avk::push_constant_binding_data{
            avk::shader_type::vertex | avk::shader_type::fragment, 0,
            sizeof(int) },

        avk::descriptor_binding(0, 0,
            avk::as_combined_image_samplers(
                mScene.mImageSamplers, avk::layout::shader_read_only_optimal)),
        avk::descriptor_binding(0, 1, mViewProjBuffer),
        avk::descriptor_binding(0, 2, mModelMatrices),
        avk::descriptor_binding(1, 0, mScene.mMaterialBuffer));

    /*********************************************
    LIGHTING
    *********************************************/

    auto lighingRenderPass = avk::context().create_renderpass(
        { avk::attachment::declare(avk::format_from_window_color_buffer(
                                       avk::context().main_window()),
              avk::on_load::clear.from_previous_layout(avk::layout::undefined),
              avk::usage::color(0),
              avk::on_store::store.in_layout(avk::layout::present_src))
                .set_clear_color(
                    CLEAR_COLOR), // ONLY THIS IS SHOWN, TODO: FIND OUT WHY

            avk::attachment::declare(avk::format_from_window_depth_buffer(
                                         avk::context().main_window()),
                avk::on_load::clear.from_previous_layout(
                    avk::layout::undefined),
                avk::usage::depth_stencil, avk::on_store::dont_care)

        },
        avk::context()
            .main_window()
            ->renderpass_reference()
            .subpass_dependencies());

    mLightingPipeline = avk::context().create_graphics_pipeline_for(
        avk::vertex_shader("shaders/def_lighting.vert"),
        avk::fragment_shader("shaders/def_lighting.frag"),
        avk::cfg::culling_mode::disabled,
        avk::cfg::viewport_depth_scissors_config::from_framebuffer(
            avk::context().main_window()->backbuffer_reference_at_index(0)),
        lighingRenderPass,
        // Input for shaders (here all for fragment)
        avk::push_constant_binding_data{
            avk::shader_type::fragment, 0, sizeof(mPCsLightPass) },
        avk::descriptor_binding(0, 0,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[0] },
                avk::layout::shader_read_only_optimal)), // gPosition
        avk::descriptor_binding(0, 1,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[1] },
                avk::layout::shader_read_only_optimal)), // gAlbedo
        avk::descriptor_binding(0, 2,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[2] },
                avk::layout::shader_read_only_optimal)), // gNormal
        avk::descriptor_binding(0, 3,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[3] },
                avk::layout::depth_stencil_read_only_optimal)), // gDepth

        avk::descriptor_binding(0, 4,
            avk::as_combined_image_samplers(
                std::vector{
                    mVolumetricLightingImageSamplers[0] }, // Vol Light Texture
                avk::layout::shader_read_only_optimal)),
        avk::descriptor_binding(0, 5, mLightBuffer),
        avk::descriptor_binding(0, 6,
            avk::as_combined_image_samplers(mShadows.mShadowMapSamplers,
                avk::layout::shader_read_only_optimal)),
        avk::descriptor_binding(0, 7, mShadows.mLightVPsBuffer));

    /*********************************************
    VOLUMETRIC
    *********************************************/

    // pipeline
    mVolumetricLightPipeline = avk::context().create_graphics_pipeline_for(
        // Shader stages:
        avk::vertex_shader("shaders/volumetric_light.vert"),
        avk::fragment_shader("shaders/volumetric_light.frag"),
        avk::cfg::culling_mode::disabled,

        // Fullscreen quad doesn't require per-vertex input; omit bindings:
        // Alternatively, you can use fullscreen quad generation in the vertex
        // shader.

        // Render state configurations:
        avk::cfg::front_face::define_front_faces_to_be_counter_clockwise(),
        avk::cfg::viewport_depth_scissors_config::from_framebuffer(
            mVolumetricLightFramebuffer.as_reference()),

        // Attachments (single output for volumetric light contribution):
        volumetricLightAttachment, // Output texture for volumetric light
                                   // contribution

        // Descriptor bindings:
        // 1. Binding for accessing G-Buffer textures (position, albedo,
        // normal)
        /*
        avk::descriptor_binding(0, 0,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[0] },
                avk::layout::shader_read_only_optimal)), // gPosition
        */
        avk::descriptor_binding(0, 1,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[1] },
                avk::layout::shader_read_only_optimal)), // gAlbedo
        avk::descriptor_binding(0, 2,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[2] },
                avk::layout::shader_read_only_optimal)), // gNormal
        avk::descriptor_binding(0, 0,
            avk::as_combined_image_samplers(
                std::vector{ mGBufferImageSamplers[3] },
                avk::layout::depth_stencil_read_only_optimal)), // gDepth
        // 2. Binding for additional inputs like light properties, camera
        // (push_constant) matrices, inverse matrix:
        avk::descriptor_binding(0, 3, mInvViewProjBuffer), // Light position
        avk::descriptor_binding(
            0, 4, mVolumetricLightBuffer), // Inverse view-projection matrix
        avk::push_constant_binding_data{ avk::shader_type::fragment, 0,
            sizeof(glm::vec3) } // for camera position

    );
    // set up updater
    // we want to use an updater, so create one:
    mUpdater.emplace();
    mLightingPipeline
        .enable_shared_ownership(); // Make it usable with the updater
    mGBufferPipeline.enable_shared_ownership();
    mVolumetricLightPipeline.enable_shared_ownership();
    // set speeds of cameras to match scaling of objects. Change this to some
    // number that feels right.
    // TODO: check what works
    auto speedMultiplier = 3.5f; // mScale.x;
    auto rotationSpeedMultiplier = 1.0f;
    mQuakeCam.set_move_speed(mQuakeCam.move_speed() * speedMultiplier);
    mQuakeCam.set_rotation_speed(
        mQuakeCam.rotation_speed() * rotationSpeedMultiplier);

    // Add the cameras to the composition (and let them handle updates)
    mQuakeCam.set_translation({ 2.0f, 3.0f, -2.0f });
    mQuakeCam.set_perspective_projection(glm::radians(60.0f),
        avk::context().main_window()->aspect_ratio(), 0.3f, 1000.0f);
    avk::current_composition()->add_element(mQuakeCam);

    geo_has_finished = avk::context().create_semaphore();
    geo_has_finished.enable_shared_ownership();

    rendering_has_finished = avk::context().create_semaphore();
    rendering_has_finished.enable_shared_ownership();

    volumetricLighting_has_finished = avk::context().create_semaphore();
    volumetricLighting_has_finished.enable_shared_ownership();
  }

  void render() override
  {
    auto mainWnd = avk::context().main_window();

    auto viewProjMat = mQuakeCam.projection_and_view_matrix();
    auto emptyCmd = mViewProjBuffer->fill(glm::value_ptr(viewProjMat), 0);

    auto invViewProjMat = glm::inverse(viewProjMat);
    emptyCmd = mInvViewProjBuffer->fill(glm::value_ptr(invViewProjMat), 0);

    auto const modelMatrix = glm::mat4(1.0);
    auto const normalMatrix = glm::transpose(glm::inverse(modelMatrix));

    // Get a command pool to allocate command buffers from:
    auto &commandPool
        = avk::context().get_command_pool_for_single_use_command_buffers(
            *mQueue);

    // The swap chain provides us with an "image available semaphore" for the
    // current frame. Only after the swapchain image has become available, we
    // may start rendering into it.
    auto imageAvailableSemaphore
        = mainWnd->consume_current_image_available_semaphore();

    // Create command buffers
    auto cmdBfrs = commandPool->alloc_command_buffers(
        3, vk::CommandBufferUsageFlagBits::eOneTimeSubmit);

    /*********************************************
    GEOMETRY
    *********************************************/
    avk::context()
        .record({ avk::command::render_pass(
            mGBufferPipeline->renderpass_reference(),
            mGBufferFramebuffer.as_reference(),
            {
                avk::command::bind_pipeline(mGBufferPipeline.as_reference()),
                avk::command::bind_descriptors(mGBufferPipeline->layout(),
                    mDescriptorCache->get_or_create_descriptor_sets(
                        { avk::descriptor_binding(0, 0,
                              avk::as_combined_image_samplers(
                                  mScene.mImageSamplers,
                                  avk::layout::shader_read_only_optimal)),
                            avk::descriptor_binding(0, 1, mViewProjBuffer),
                            avk::descriptor_binding(0, 2, mModelMatrices),
                            avk::descriptor_binding(
                                1, 0, mScene.mMaterialBuffer) })),

                // Draw all the draw calls:
                avk::command::custom_commands(
                    [&, this](avk::command_buffer_t &cb)
                    {
                      for (auto &drawCall : mScene.mDrawCalls)
                      {
                        cb.record({ avk::command::push_constants(
                                        mGBufferPipeline->layout(),
                                        drawCall.mMaterialIndex),

                            avk::command::draw_indexed(
                                drawCall.mIndexBuffer.as_reference(),
                                drawCall.mPositionsBuffer.as_reference(),
                                drawCall.mTexCoordsBuffer.as_reference(),
                                drawCall.mNormalsBuffer.as_reference(),
                                drawCall.mTagentsBuffer.as_reference(),
                                drawCall.mBitangentsBuffer.as_reference()) });
                      }
                    }),
            }) })
        .into_command_buffer(cmdBfrs[0])
        .then_submit_to(*mQueue)
        .waiting_for(
            imageAvailableSemaphore >> avk::stage::color_attachment_output)
        .signaling_upon_completion(
            avk::stage::geometry_shader >> geo_has_finished)
        .submit();

    /*********************************************
    VOLUMETRIC
    *********************************************/
    // set camerapostion and inverse matrix for VolLightData
    // do all the normal record stuff
    // add texture output to lighting pass

    avk::context()
        .record({ avk::command::render_pass(
            mVolumetricLightPipeline->renderpass_reference(),
            mVolumetricLightFramebuffer
                .as_reference(), // Framebuffer for volumetric light pass
            {
                avk::command::bind_pipeline(
                    mVolumetricLightPipeline.as_reference()),

                // Bind the descriptors for the volumetric light pass
                avk::command::bind_descriptors(
                    mVolumetricLightPipeline->layout(),
                    mDescriptorCache->get_or_create_descriptor_sets({
                        avk::descriptor_binding(0, 0,
                            avk::as_combined_image_samplers(
                                std::vector{ mGBufferImageSamplers[3] },
                                avk::layout::attachment_optimal)), // gDepth
                        avk::descriptor_binding(0, 1,
                            avk::as_combined_image_samplers(
                                std::vector{ mGBufferImageSamplers[1] },
                                avk::layout::attachment_optimal)), // gAlbedo
                        avk::descriptor_binding(0, 2,
                            avk::as_combined_image_samplers(
                                std::vector{ mGBufferImageSamplers[2] },
                                avk::layout::attachment_optimal)), // gNormal
                        avk::descriptor_binding(0, 3,
                            mInvViewProjBuffer), // Inverse view-projection
                                                 // matrix
                        avk::descriptor_binding(
                            0, 4, mVolumetricLightBuffer), // Volumetric light
                                                           // data buffer
                    })),

                // Custom commands to run the volumetric light rendering
                avk::command::custom_commands(
                    [&, this](avk::command_buffer_t &cb)
                    {
                      cb.record(
                          { avk::command::push_constants(
                                mVolumetricLightPipeline->layout(),
                                glm::vec3(
                                    mQuakeCam.translation()), // Pass camera
                                                              // position as
                                                              // push constant
                                avk::shader_type::fragment),

                              // Draw the fullscreen quad for the volumetric
                              // light pass
                              avk::command::draw(3, 1, 0, 0) });
                    }),
            }) })
        .into_command_buffer(
            cmdBfrs[1]) // Command buffer for volumetric lighting pass
        .then_submit_to(*mQueue)
        .waiting_for(geo_has_finished >> avk::stage::geometry_shader)
        .signaling_upon_completion(
            avk::stage::all_graphics >> volumetricLighting_has_finished)
        .submit();

    /*********************************************
    LIGHTING
    *********************************************/

    mPCsLightPass.cameraPosition = mQuakeCam.translation();

    avk::context()
        .record({ avk::command::render_pass(
            mLightingPipeline->renderpass_reference(),
            avk::context().main_window()->current_backbuffer_reference(),
            {
                avk::command::bind_pipeline(mLightingPipeline.as_reference()),
                avk::command::bind_descriptors(mLightingPipeline->layout(),
                    mDescriptorCache->get_or_create_descriptor_sets(
                        { avk::descriptor_binding(0, 0,
                              avk::as_combined_image_samplers(
                                  std::vector{ mGBufferImageSamplers[0] },
                                  avk::layout::
                                      attachment_optimal)), // gPosition

                            avk::descriptor_binding(0, 1,
                                avk::as_combined_image_samplers(
                                    std::vector{ mGBufferImageSamplers[1] },
                                    avk::layout::
                                        attachment_optimal)), // gAlbedo

                            avk::descriptor_binding(0, 2,
                                avk::as_combined_image_samplers(
                                    std::vector{ mGBufferImageSamplers[2] },
                                    avk::layout::
                                        attachment_optimal)), // gNormal

                            avk::descriptor_binding(0, 3,
                                avk::as_combined_image_samplers(
                                    std::vector{ mGBufferImageSamplers[3] },
                                    avk::layout::
                                        attachment_optimal)), // gDepth
                            avk::descriptor_binding(0, 4,
                                avk::as_combined_image_samplers(
                                    std::vector{
                                        mVolumetricLightingImageSamplers
                                            [0] }, // vol light
                                    avk::layout::attachment_optimal)),
                            avk::descriptor_binding(
                                0, 5, mLightBuffer), // lights
                            avk::descriptor_binding(0, 6,
                                avk::as_combined_image_samplers(
                                    mShadows.mShadowMapSamplers,
                                    avk::layout::attachment_optimal)),
                            avk::descriptor_binding(
                                0, 7, mShadows.mLightVPsBuffer) })),

                avk::command::custom_commands(
                    [&, this](avk::command_buffer_t &cb)
                    {
                      cb.record(
                          { avk::command::push_constants(
                                mLightingPipeline->layout(), mPCsLightPass,
                                avk::shader_type::fragment),
                              avk::command::draw(3, 1, 0, 0) });
                    }),
            }) })
        .into_command_buffer(cmdBfrs[2])
        .then_submit_to(*mQueue)
        .waiting_for(
            volumetricLighting_has_finished >> avk::stage::all_graphics)
        .signaling_upon_completion(
            avk::stage::color_attachment_output >> rendering_has_finished)
        // this was previously done by the ui layer since it was the last pass
        .signaling_upon_completion(mainWnd->use_current_frame_finished_fence())
        .submit();

    mainWnd->add_present_dependency_for_current_frame(rendering_has_finished);

    for (size_t i = 0; i < cmdBfrs.size(); ++i)
      mainWnd->handle_lifetime(std::move(cmdBfrs[i]));
  }

  void update() override
  {
    static int counter = 0;
    if (++counter == 4)
    {
      auto current = std::chrono::high_resolution_clock::now();
      auto time_span = current - mInitTime;
      auto int_min
          = std::chrono::duration_cast<std::chrono::minutes>(time_span)
                .count();
      auto int_sec
          = std::chrono::duration_cast<std::chrono::seconds>(time_span)
                .count();
      auto fp_ms
          = std::chrono::duration<double, std::milli>(time_span).count();
      printf("Time from init to fourth frame: %d min, %lld sec %lf ms\n",
          int_min, int_sec - static_cast<decltype(int_sec)>(int_min) * 60,
          fp_ms - 1000.0 * int_sec);
    }

    if (avk::input().key_pressed(avk::key_code::escape)
        || avk::context().main_window()->should_be_closed())
    {
      // Stop the current composition:
      avk::current_composition()->stop();
    }
    if (avk::input().key_pressed(avk::key_code::left))
    {
      mQuakeCam.look_along(avk::left());
    }
    if (avk::input().key_pressed(avk::key_code::right))
    {
      mQuakeCam.look_along(avk::right());
    }
    if (avk::input().key_pressed(avk::key_code::up))
    {
      mQuakeCam.look_along(avk::front());
    }
    if (avk::input().key_pressed(avk::key_code::down))
    {
      mQuakeCam.look_along(avk::back());
    }
    if (avk::input().key_pressed(avk::key_code::page_up))
    {
      mQuakeCam.look_along(avk::up());
    }
    if (avk::input().key_pressed(avk::key_code::page_down))
    {
      mQuakeCam.look_along(avk::down());
    }
    if (avk::input().key_pressed(avk::key_code::home))
    {
      mQuakeCam.look_at(glm::vec3{ 0.0f, 0.0f, 0.0f });
    }
    if (avk::input().key_pressed(avk::key_code::num0))
    {
      mPCsLightPass.debugMode = 0;
    }
    if (avk::input().key_pressed(avk::key_code::num1))
    {
      mPCsLightPass.debugMode = 1;
    }
    if (avk::input().key_pressed(avk::key_code::num2))
    {
      mPCsLightPass.debugMode = 2;
    }
    if (avk::input().key_pressed(avk::key_code::num3))
    {
      mPCsLightPass.debugMode = 3;
    }
    if (avk::input().key_pressed(avk::key_code::num4))
    {
      mPCsLightPass.debugMode = 4;
    }
    if (avk::input().key_pressed(avk::key_code::num5))
    {
      mPCsLightPass.debugMode = 5;
    }
    if (avk::input().key_pressed(avk::key_code::num6))
    {
      mPCsLightPass.debugMode = 6;
    }
    // camera
    if (avk::input().key_pressed(avk::key_code::f1))
    {
      if (mCameraPathEnabled)
      { // => disable
        if (mCameraPath.has_value())
        {
          avk::current_composition()->remove_element_immediately(
              mCameraPath.value());
          mCameraPath.reset();
          mCameraPathEnabled = false;
        }
      }
      else
      { // => enable
        if (mCameraPath.has_value())
        {
          avk::current_composition()->remove_element_immediately(
              mCameraPath.value());
        }
        mCameraPath.emplace(mQuakeCam);
        avk::current_composition()->add_element(mCameraPath.value());
        mCameraPathEnabled = true;
      }
    }
  }

private: // v== Member variables ==v
  std::chrono::high_resolution_clock::time_point mInitTime;

  avk::queue *mQueue;
  avk::descriptor_cache mDescriptorCache;

  avk::buffer mViewProjBuffer;
  avk::buffer mLightBuffer;
  avk::buffer mInvViewProjBuffer;
  avk::buffer mVolumetricLightBuffer;

  avk::buffer mModelMatrices;

  std::vector<avk::image_sampler> mGBufferImageSamplers;
  std::vector<avk::image_sampler> mVolumetricLightingImageSamplers;

  avk::semaphore rendering_has_finished;
  avk::semaphore geo_has_finished;
  avk::semaphore volumetricLighting_has_finished;

  Scene mScene;
  avk::graphics_pipeline mGBufferPipeline;
  avk::graphics_pipeline mLightingPipeline;
  avk::graphics_pipeline mVolumetricLightPipeline;
  glm::vec3 mScale;

  avk::image_view mGBufferPositionView;
  avk::image_view mGBufferNormalView;
  avk::image_view mGBufferAlbedoView;
  avk::image_view mGBufferDepthView;
  avk::framebuffer mGBufferFramebuffer;

  avk::image_view mVolumetricLightView;
  avk::framebuffer mVolumetricLightFramebuffer;

  Shadows mShadows;

  avk::quake_camera mQuakeCam;
  std::optional<camera_path> mCameraPath;
  bool mCameraPathEnabled = true;

  PushConstantsLightPass mPCsLightPass
      = { int(0), glm::vec3(0.0f, 0.0f, 0.0f) };

  // clang-format off
  LightData mlightData = {
    // Positions
    {
        { 10.0f, 3.0f, 0.0f, 1.0f },  // test lights - deactivated via radius = 0
        { -10.0f, 3.0f, 0.0f, 1.0f },
        { 0.0f, 3.0f, 5.0f, 1.0f },
        { -8.0f, 1.0f, 1.0f, 1.0f },
        { 8.0f, 15.0f, 22.0f, 1.0f },  // general light Position of light 4 (spot light)

        { 1.5f, 10.0f, 16.5f, 1.0f },  // lights above paintings
        { 14.0f, 10.0f, 19.0f, 1.0f },
        { 18.0f, 10.0f, 19.0f, 1.0f },
        { 3.5f, 10.0f, 24.5f, 1.0f },
        { -0.5f, 10.0f, 24.5f, 1.0f },
        { 13.0f, 10.0f, 27.5f, 1.0f },

        { 6.5f, 5.0f, 16.5f, 1.0f },   //lights towards paintings
        { 13.8f, 5.0f, 24.0f, 1.0f },
        { 18.1f, 5.0f, 24.0f, 1.0f },
        { 3.6f, 5.0f, 29.5f, 1.0f },
        { -0.6f, 5.0f, 29.5f, 1.0f },
        { 8.0f, 5.0f, 27.5f, 1.0f },

        { 7.5f, 15.0f, 7.5f, 1.0f },    // room 2
        { -7.0f, 15.0f, -7.5f, 1.0f },
        { 7.5f, 15.0f, -7.5f, 1.0f },
        { -7.5f, 15.0f, 7.5f, 1.0f },

        { 18.3f, 10.0f, 8.0f, 1.0f },   //bunny
        { 18.3f, 10.0f, -8.0f, 1.0f },
        { -18.0f, 10.0f, 0.5f, 1.0f },

    },
    
    // Colors
    {
        { 1.0f, 0.0f, 0.0f, 1.0f },  // Color of light 0 (red)
        { 0.0f, 1.0f, 0.0f, 1.0f },  // Color of light 1 (green)
        { 0.0f, 0.0f, 1.0f, 1.0f },  // Color of light 2 (blue)
        { 1.0f, 0.0f, 1.0f, 1.0f },   // Color of light 3 (purple)
        { 1.0f, 1.0f, 1.0f, 1.0f },  // general light Color of light 4 (white)

        { 1.0f, 1.0f, 0.0f, 1.0f },  // lights above paintings
        { 1.0f, 1.0f, 0.0f, 1.0f },
        { 1.0f, 1.0f, 0.0f, 1.0f },
        { 1.0f, 1.0f, 0.0f, 1.0f },
        { 1.0f, 1.0f, 0.0f, 1.0f },
        { 1.0f, 1.0f, 0.0f, 1.0f },

        { 0.5f, 0.5f, 1.0f, 1.0f }, //lights towards paintings
        { 0.5f, 0.5f, 1.0f, 1.0f },
        { 0.5f, 0.5f, 1.0f, 1.0f },
        { 0.5f, 0.5f, 1.0f, 1.0f },
        { 0.5f, 0.5f, 1.0f, 1.0f },
        { 0.5f, 0.5f, 1.0f, 1.0f },

        { 0.25f, 0.0f, 0.0f, 1.0f },  //room 2
        { 0.0f, 0.25f, 0.0f, 1.0f }, 
        { 0.0f, 0.0f, 0.25f, 1.0f },
        { 0.0833f, 0.0833f, 0.0833f, 1.0f },

        { 1.0f, 0.0f, 0.0f, 1.0f },  //bunny
        { 1.0f, 0.0f, 0.0f, 1.0f },
        { 1.0f, 0.0f, 0.0f, 1.0f },
     },

    // Targets for spotlights (if top-down just same as positions but with y=0)
    {
        { 10.0f, 0.0f, 0.0f, 1.0f },  // test lights
        { -10.0f, 0.0f, 0.0f, 1.0f },
        { 0.0f, 0.0f, 5.0f, 1.0f },
        { -8.0f, 0.0f, 1.0f, 1.0f },
        { 8.0f, 0.0f, 22.0f, 1.0f }, // general light

        { 1.5f, 0.0f, 16.5f, 1.0f },  // lights above paintings
        { 14.0f, 0.0f, 19.0f, 1.0f },
        { 18.0f, 0.0f, 19.0f, 1.0f },
        { 3.5f, 0.0f, 24.5f, 1.0f },
        { -0.5f, 0.0f, 24.5f, 1.0f },
        { 13.0f, 0.0f, 27.5f, 1.0f },

        { 1.5f, 2.5f, 16.5f, 1.0f }, // lights towards paintings
        { 13.8f, 2.5f, 19.0f, 1.0f },
        { 18.1f, 2.5f, 19.0f, 1.0f },
        { 3.6f, 2.5f, 24.5f, 1.0f },
        { -0.6f, 2.5f, 24.5f, 1.0f },
        { 13.0f, 2.5f, 27.5f, 1.0f },

        { 0.0f, 0.0f, 0.0f, 1.0f }, // room 2
        { 0.0f, 0.0f, 0.0f, 1.0f },
        { 0.0f, 0.0f, 0.0f, 1.0f },
        { 0.0f, 0.0f, 0.0f, 1.0f },

        { 18.3f, 0.0f, 8.0f, 1.0f }, // bunny
        { 18.3f, 0.0f, -8.0f, 1.0f },
        { -18.0f, 0.0f, 0.5f, 1.0f },
     },
    
    // Radius (attenuation radius for point lights)
    {
        { 00.0f, 0.0f, 0.0f, 0.0f }, // Radius for light 0
        { 00, 0.0f, 0.0f, 0.0f }, // Radius for light 1
        { 00.0f, 0.0f, 0.0f, 0.0f }, // Radius for light 2
        { 00.0f, 0.0f, 0.0f, 0.0f },   // Radius for light 3
        { 25.0f, 0.0f, 0.0f, 0.0f },  //general light

        { 15.0f, 0.0f, 0.0f, 0.0f },  //lights above paintings
        { 15.0f, 0.0f, 0.0f, 0.0f },
        { 15.0f, 0.0f, 0.0f, 0.0f },
        { 15.0f, 0.0f, 0.0f, 0.0f },
        { 15.0f, 0.0f, 0.0f, 0.0f },
        { 15.0f, 0.0f, 0.0f, 0.0f },

        { 10.0f, 0.0f, 0.0f, 0.0f },    // lights towards paintings
        { 10.0f, 0.0f, 0.0f, 0.0f },
        { 10.0f, 0.0f, 0.0f, 0.0f },
        { 10.0f, 0.0f, 0.0f, 0.0f },
        { 10.0f, 0.0f, 0.0f, 0.0f },
        { 10.0f, 0.0f, 0.0f, 0.0f },

        { 30.0f, 0.0f, 0.0f, 0.0f }, // room 2
        { 30.0f, 0.0f, 0.0f, 0.0f },
        { 30.0f, 0.0f, 0.0f, 0.0f },
        { 30.0f, 0.0f, 0.0f, 0.0f },

        { 15.0f, 0.0f, 0.0f, 0.0f },
        { 15.0f, 0.0f, 0.0f, 0.0f },
        { 15.0f, 0.0f, 0.0f, 0.0f },
    },
    
    // Cone Angles
    {
        { glm::radians(40.0f), glm::radians(50.0f), 0.0f, 0.0f },  // Cone angle for light 0 (spotlight)
        { glm::radians(40.0f), glm::radians(50.0f), 0.0f, 0.0f },   // Cone angle for light 1 (spotlight)
        { glm::radians(40.0f), glm::radians(50.0f), 0.0f, 0.0f },    // Cone angle for light 2 (spotlight)
        { glm::radians(40.0f), glm::radians(50.0f), 0.0f, 0.0f },  // Cone angle for light 3 (spotlight)
        { glm::radians(60.0f), glm::radians(80.0f), 0.0f, 0.0f },   //general light

        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },   //lights towards paintings
        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },
        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },
        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },
        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },
        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },

        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },   // lights towards paintings
        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },
        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },
        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },
        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },
        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },

        { glm::radians(20.0f), glm::radians(25.0f), 0.0f, 0.0f },   // room 2
        { glm::radians(20.0f), glm::radians(25.0f), 0.0f, 0.0f },
        { glm::radians(20.0f), glm::radians(25.0f), 0.0f, 0.0f },
        { glm::radians(20.0f), glm::radians(25.0f), 0.0f, 0.0f },

        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },   //bunny
        { glm::radians(20.0f), glm::radians(30.0f), 0.0f, 0.0f },
        { glm::radians(10.0f), glm::radians(20.0f), 0.0f, 0.0f },

    }
    // clang-format on
  };

  VolumetricLightData mVolumetricLightData = {
    { -14.0f, 2.0f, 15.0f, 0.0f },
    glm::vec4(),
    { 256,    0.8,  0.0,   0.0  }  //  x for sample number, y for scattering factor, z
  // for unused, w for unused
  };

}; // model_loader_app

int main() // <== Starting point ==
{
  int result = EXIT_FAILURE;
  try
  {
    // Create a window and open it
    auto mainWnd = avk::context().create_window("rtr24: CGMuseum");

    mainWnd->set_resolution({ 1920, 1080 });
    mainWnd->enable_resizing(false);
    mainWnd->set_additional_back_buffer_attachments(
        { avk::attachment::declare(vk::Format::eD32Sfloat,
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::depth_stencil, avk::on_store::dont_care) });
    mainWnd->set_presentaton_mode(avk::presentation_mode::fifo);
    mainWnd->open();

    auto &singleQueue = avk::context().create_queue(
        {}, avk::queue_selection_preference::versatile_queue, mainWnd);
    mainWnd->set_queue_family_ownership(singleQueue.family_index());
    mainWnd->set_present_queue(singleQueue);

    // Create an instance of our main avk::element which contains all the
    // functionality:
    auto app = Demo(singleQueue);

    // Initialize OpenAL for audio
    ALCdevice *device = nullptr;
    ALCcontext *context = nullptr;
    app.initializeOpenAL(device, context);

    // Load and play background music using app
    auto audioData = app.loadAudioFile(
        "audio_assets//Beethoven_Moonlight_Sonata_1st_Movement.wav");
    app.startAudioPlayback(app, audioData);

    // Compile all the configuration parameters and the invokees into a
    // "composition":
    auto composition = configure_and_compose(
        avk::application_name("rtr24: CGMuseum"),
        [](avk::validation_layers &config)
        {
          config.enable_feature(
              vk::ValidationFeatureEnableEXT::eSynchronizationValidation);
        },
        // Pass windows:
        mainWnd,
        // Pass invokees:
        app);

    // Create an invoker object, which defines the way how invokees/elements
    // are invoked (In this case, just sequentially in their execution order):
    avk::sequential_invoker invoker;

    // With everything configured, let us start our render loop:
    composition.start_render_loop(
        // Callback in the case of update:
        [&invoker](const std::vector<avk::invokee *> &aToBeInvoked)
        {
          // Call all the update() callbacks:
          invoker.invoke_updates(aToBeInvoked);
        },
        // Callback in the case of render:
        [&invoker](const std::vector<avk::invokee *> &aToBeInvoked)
        {
          // Sync (wait for fences and so) per window BEFORE executing render
          // callbacks
          avk::context().execute_for_each_window(
              [](avk::window *wnd) { wnd->sync_before_render(); });

          // Call all the render() callbacks:
          invoker.invoke_renders(aToBeInvoked);

          // Render per window:
          avk::context().execute_for_each_window(
              [](avk::window *wnd) { wnd->render_frame(); });
        }); // This is a blocking call, which loops until
            // avk::current_composition()->stop(); has been called (see
            // update())

    app.cleanupOpenAL(device, context);

    result = EXIT_SUCCESS;
  }
  catch (avk::logic_error &)
  {
  }
  catch (avk::runtime_error &)
  {
  }
  return result;
}