#include "scene.h"

#include <model.hpp>
#include <context_vulkan.hpp>
#include <material_image_helpers.hpp>

Scene Scene::load(std::string path, avk::queue *queue)
{
  Scene scene;

  auto sceneData = avk::model_t::load_from_file(
      path, aiProcess_Triangulate | aiProcess_PreTransformVertices
                | aiProcess_CalcTangentSpace);
  auto const distinctMaterials = sceneData->distinct_material_configs();

  std::vector<avk::material_config> allMaterialsConfigs;
  allMaterialsConfigs.reserve(distinctMaterials.size());
  for (auto const &[config, materialIndices] : distinctMaterials)
  {
    allMaterialsConfigs.push_back(config);
    auto const materialIndex
        = static_cast<int>(allMaterialsConfigs.size() - 1);

    for (auto index : materialIndices)
    {
      auto &drawCall = scene.mDrawCalls.emplace_back();

      drawCall.mMaterialIndex = materialIndex;

      avk::append_indices_and_vertex_data(
          avk::additional_index_data(drawCall.mIndices,
              [&] { return sceneData->indices_for_mesh<uint32_t>(index); }),
          avk::additional_vertex_data(drawCall.mPositions,
              [&] { return sceneData->positions_for_mesh(index); }),
          avk::additional_vertex_data(drawCall.mTexCoords,
              [&] {
                return sceneData->texture_coordinates_for_mesh<glm::vec2>(
                    index, 0);
              }),
          avk::additional_vertex_data(drawCall.mNormals,
              [&] { return sceneData->normals_for_mesh(index); }),
          avk::additional_vertex_data(drawCall.mTagents,
              [&] { return sceneData->tangents_for_mesh(index); }),
          avk::additional_vertex_data(drawCall.mBitangents,
              [&] { return sceneData->bitangents_for_mesh(index); }));

      drawCall.mPositionsBuffer
          = avk::context().create_buffer(avk::memory_usage::device, {},
              avk::vertex_buffer_meta::create_from_data(drawCall.mPositions));
      auto posFillCmd
          = drawCall.mPositionsBuffer->fill(drawCall.mPositions.data(), 0);

      drawCall.mTexCoordsBuffer
          = avk::context().create_buffer(avk::memory_usage::device, {},
              avk::vertex_buffer_meta::create_from_data(drawCall.mTexCoords));
      auto tcoFillCmd
          = drawCall.mTexCoordsBuffer->fill(drawCall.mTexCoords.data(), 0);

      drawCall.mNormalsBuffer
          = avk::context().create_buffer(avk::memory_usage::device, {},
              avk::vertex_buffer_meta::create_from_data(drawCall.mNormals));
      auto nrmFillCmd
          = drawCall.mNormalsBuffer->fill(drawCall.mNormals.data(), 0);

      drawCall.mTagentsBuffer
          = avk::context().create_buffer(avk::memory_usage::device, {},
              avk::vertex_buffer_meta::create_from_data(drawCall.mTagents));
      auto tngFillCmd
          = drawCall.mTagentsBuffer->fill(drawCall.mTagents.data(), 0);

      drawCall.mBitangentsBuffer
          = avk::context().create_buffer(avk::memory_usage::device, {},
              avk::vertex_buffer_meta::create_from_data(drawCall.mBitangents));
      auto btngFillCmd
          = drawCall.mBitangentsBuffer->fill(drawCall.mBitangents.data(), 0);

      drawCall.mIndexBuffer
          = avk::context().create_buffer(avk::memory_usage::device, {},
              avk::index_buffer_meta::create_from_data(drawCall.mIndices));
      auto idxFillCmd
          = drawCall.mIndexBuffer->fill(drawCall.mIndices.data(), 0);

      auto fence = avk::context().record_and_submit_with_fence(
          { std::move(posFillCmd), std::move(tcoFillCmd),
              std::move(nrmFillCmd), std::move(tngFillCmd),
              std::move(btngFillCmd), std::move(idxFillCmd) },
          *queue);
      fence->wait_until_signalled();
    }
  }

  auto [gpuMaterials, imageSamplers, materialCommands]
      = avk::convert_for_gpu_usage<avk::material_gpu_data>(allMaterialsConfigs,
          false, true, avk::image_usage::general_texture,
          avk::filter_mode::trilinear);

  scene.mImageSamplers.swap(imageSamplers);

  scene.mMaterialBuffer
      = avk::context().create_buffer(avk::memory_usage::device, {},
          avk::storage_buffer_meta::create_from_data(gpuMaterials));

  auto matFence = avk::context().record_and_submit_with_fence(
      { std::move(materialCommands),
          scene.mMaterialBuffer->fill(gpuMaterials.data(), 0) },
      *queue);
  matFence->wait_until_signalled();

  return scene;
}