#version 450
#extension GL_EXT_debug_printf: enable
#extension GL_EXT_nonuniform_qualifier: require

#define NUM_LIGHTS 24

layout (location = 0) out vec4 outColor;// Final output color

layout (binding = 0) uniform sampler2D gPosition;// G-Buffer Position
layout (binding = 1) uniform sampler2D gAlbedo;// G-Buffer Albedo
layout (binding = 2) uniform sampler2D gNormal;// G-Buffer Normal
layout (binding = 3) uniform sampler2D gDepth;// G-Buffer Depth
layout (binding = 4) uniform sampler2D volLight;// G-Buffer Depth


layout (std140, binding = 5) uniform LightData {
    vec4 lightPositions[NUM_LIGHTS];
    vec4 lightColors[NUM_LIGHTS];
    vec4 lightTargets[NUM_LIGHTS];
    vec4 radius[NUM_LIGHTS];
    vec4 coneAngles[NUM_LIGHTS];
} lightData;

layout (binding = 6) uniform sampler2D shadowMaps[];

layout (binding = 7) uniform LightVPs {
    mat4 m[NUM_LIGHTS];
} lightVPs;

layout (push_constant) uniform PushConstants {
    int debugMode;//0 = standard rendering, 1 = pos, 2 = albedo, 3 = normals, 4 = depth
    vec3 cameraPosition;
} pushConstants;

float computeShadow(vec3 normal, vec3 lightDir, vec4 positionLS, int shadowMapIndex)
{
    vec3 projectedPos = positionLS.xyz / positionLS.w;
    if (projectedPos.z > 1.0) return 0.0;
    float depthMin = texture(shadowMaps[shadowMapIndex], projectedPos.xy).r;
    float fragDepth = projectedPos.z;
    float bias = max(0.05 * (1.0 - dot(normal, lightDir)), 0.001);
    float shadow = 0.0;
    vec2 texelSize = 1.0 / textureSize(shadowMaps[shadowMapIndex], 0);
    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            float pcfDepth = texture(shadowMaps[shadowMapIndex], projectedPos.xy + vec2(x, y) * texelSize).r;
            shadow += fragDepth - bias > pcfDepth ? 1.0 : 0.0;
        }
    }
    shadow /= 9.0;
    return 1 - shadow;
}

void main() {
    // Sample G-Buffer
    //normalize to uv coordinates [0,1] then texture() of the individual ones
    vec3 position = texture(gPosition, gl_FragCoord.xy / vec2(textureSize(gPosition, 0))).rgb;
    vec3 normal = texture(gNormal, gl_FragCoord.xy / vec2(textureSize(gNormal, 0))).rgb;
    vec3 albedo = texture(gAlbedo, gl_FragCoord.xy / vec2(textureSize(gAlbedo, 0))).rgb;
    float depth = texture(gDepth, gl_FragCoord.xy / vec2(textureSize(gDepth, 0))).r;
    vec3 volumetric = texture(volLight, gl_FragCoord.xy / vec2(textureSize(volLight, 0))).rgb;

    switch (pushConstants.debugMode) {
        case 0:
        case 6:
        vec3 finalColor = vec3(0.0);
        for (int i = 0; i < NUM_LIGHTS; i++) {
            // Spotlight calculations
            vec3 lightPos = lightData.lightPositions[i].xyz;
            vec3 lightColor = lightData.lightColors[i].rgb;
            vec3 lightTarget = lightData.lightTargets[i].xyz;// Use target instead of direction, easier to set up
            //vec3 lightDir = normalize(lightData.lightDirections[i].xyz);
            float lightRadius = lightData.radius[i].x;
            float innerCutoff = cos(lightData.coneAngles[i].x);
            float outerCutoff = cos(lightData.coneAngles[i].y);

            // Compute light direction from light position to light target
            vec3 lightDir = normalize(lightTarget - lightPos);

            // Compute light direction and distance
            vec3 toLight = lightPos - position;
            float distance = length(toLight);
            vec3 lightVec = normalize(toLight);

            // Check if within radius
            if (distance > lightRadius) {
                continue;
            }

            // Spotlight angle attenuation
            float theta = dot(lightVec, -lightDir);
            float epsilon = innerCutoff - outerCutoff;
            float intensity = clamp((theta - outerCutoff) / epsilon, 0.0, 1.0);

            // Distance attenuation
            float kl = 0.1;// Linear term
            float kq = 0.01;// Quadratic term
            float attenuation = (1.0 / (1.0 + kl * distance + kq * distance * distance)) * intensity;
            //float attenuation = (1.0 - distance / lightRadius) * intensity;

            // Diffuse shading
            float diff = max(dot(normal, lightVec), 0.0);
            vec3 diffuse = diff * lightColor * albedo;

            // Specular shading (fixed specular)
            vec3 viewDir = normalize(pushConstants.cameraPosition - position);
            vec3 halfwayDir = normalize(lightVec + viewDir);
            float spec = pow(max(dot(normal, halfwayDir), 0.0), 16.0);// Blinn-Phong specular
            vec3 specular = vec3(0.1) * spec;// Fixed specular intensity

            if (pushConstants.debugMode == 6) {
                vec4 positionLS = lightVPs.m[i] * vec4(position, 1.0);
                float shadow = computeShadow(normal, toLight, positionLS, i);

                // Accumulate light contribution
                finalColor += attenuation * (diffuse + specular) * shadow;
            } else {
                // Accumulate light contribution
                finalColor += attenuation * (diffuse + specular);
            }
        }
        finalColor += volumetric;
        outColor = vec4(finalColor, 1.0);
        break;

        case 1:// Position
        outColor = vec4(position, 1.0);
        break;

        case 2:// Normal
        outColor = vec4(normal / 2.0 + vec3(0.5), 1.0);// Map normals from [-1,1] to [0,1]
        break;

        case 3:// Albedo
        outColor = vec4(albedo, 1.0);
        break;

        case 4:// Depth
        outColor = vec4(depth, depth, depth, 1.0);
        break;

        case 5://vol light
        outColor = vec4(volumetric, 1.0);
        break;

        default :
        outColor = vec4(1.0, 0.0, 0.0, 1.0);// error screen
    }
}
