#version 450
#extension GL_EXT_debug_printf : enable

layout(location = 0) out vec4 outColor;      // Output color for the volumetric texture

// Uniforms
layout(binding = 0) uniform sampler2D gDepth;    // Depth buffer (linearized depth)
layout(binding = 1) uniform sampler2D gNormal;   // G-Buffer normals
layout(binding = 2) uniform sampler2D gAlbedo;   // G-Buffer albedo/color
layout(binding = 3) uniform Transform {
    mat4 uInvViewProj; // Inverse view-projection matrix
} transform;

layout(std140, binding = 4) uniform VolumetricLightData {
    vec4 lightPosition;  // x, y, z for position, w for unused or intensity
    vec4 lightDirection; // x, y, z for direction, w for spotlight angle (if needed)
    vec4 params;         // x for sample number, y for scattering factor, z/w unused (later tau and phi)
} volumetricLightData;

layout(push_constant) uniform PushConstants {
    vec3 cameraPosition;
} pushConstants;

// Constants
const float PI = 3.14159265359;
const float PI_RCP = 1.0 / PI;
const float tau = 0.1; // Light absorption/scattering factor
const float phi = 100.0; // Radiant power of light

// Helper function to reconstruct world position from depth
vec3 worldPositionFromDepth(float depth) {
    // Reconstruct normalized device coordinates (NDC) from gl_FragCoord
    vec2 ndc = (gl_FragCoord.xy / vec2(textureSize(gDepth, 0))) * 2.0 - 1.0;

    // Reconstruct clip-space position
    vec4 clipSpacePos = vec4(ndc, depth, 1.0);

    // Transform to world space using the inverse view-projection matrix
    vec4 worldPos = transform.uInvViewProj * clipSpacePos;

    // Divide by w to convert from homogeneous coordinates to 3D world coordinates
    return worldPos.xyz / worldPos.w;
}


// Main function
void main() {
    // Load G-buffer data
    float depth = texture(gDepth, gl_FragCoord.xy / vec2(textureSize(gDepth, 0))).r;

    // Construct world-space and light-space positions
    vec3 frag_pos = worldPositionFromDepth(depth);             // Construct world position from depth
    vec4 start_pos_worldspace = vec4(frag_pos, 1.0);           // Ray start position in WS
    vec4 end_pos_worldspace = vec4(pushConstants.cameraPosition, 1.0); // Ray end position in WS
    vec4 start_pos_lightview = volumetricLightData.lightDirection * start_pos_worldspace; // WS -> LS
    vec4 end_pos_lightview = volumetricLightData.lightDirection * end_pos_worldspace;     // WS -> LS
    vec4 delta_worldspace = normalize(end_pos_worldspace - start_pos_worldspace);         // Direction in WS
    float raymarch_distance_worldspace = length(end_pos_worldspace - start_pos_worldspace); // Distance in WS
    float step_size_worldspace = raymarch_distance_worldspace / volumetricLightData.params.x; // delta_t in WS
    vec4 ray_position_worldspace = start_pos_worldspace; // Current sample position in WS

    // Accumulate light contribution
    float light_contribution = 0.0;

    for (float l = raymarch_distance_worldspace; l > step_size_worldspace; l -= step_size_worldspace) {
        // Distance-based attenuation
        float d = length(ray_position_worldspace.xyz - volumetricLightData.lightPosition.xyz);
        float d_rcp = 1.0 / d; // 1/d for efficiency

        // Approximation of radiative transport equation without occlusion
        light_contribution += tau *
            (phi * 0.25 * PI_RCP * d_rcp * d_rcp) *
            exp(-d * tau) *
            exp(-l * tau) *
            step_size_worldspace;

        // Continue ray marching
        ray_position_worldspace += step_size_worldspace * delta_worldspace;
    }

    // Output accumulated light contribution as the result
    outColor = vec4(vec3(light_contribution), 1.0);
}
