#include "shadows.h"

#include <context_vulkan.hpp>
#include <avk/avk.hpp>
#include <cassert>

Shadows Shadows::create(std::span<glm::vec4> lightPositions,
    std::span<glm::vec4> lightTargets, glm::uvec2 screenRes, avk::queue *queue)
{
  assert(lightPositions.size() == lightTargets.size());
  auto const nLights = lightPositions.size();
  Shadows result;
  auto sampler = avk::context().create_sampler(
      avk::filter_mode::trilinear, avk::border_handling_mode::clamp_to_edge);
  std::vector<avk::attachment> attachments;

  for (size_t i = 0; i < nLights; ++i)
  {
    // up is off center because some lights point straight down (cross is 0)
    auto const lightView = glm::lookAt(glm::vec3(lightPositions[i]),
        glm::vec3(lightTargets[i]), glm::vec3(0, 1, 0.1));
    auto lightPersp = glm::perspective(glm::radians(90.f),
        float(screenRes.x) / float(screenRes.y), 0.1f, 20.f);
    // flip y coordinate, since the vulkan viewport is centered at the top left
    lightPersp[1][1] = -lightPersp[1][1];
    result.mLightVPs.push_back(lightPersp * lightView);

    result.mShadowMaps.push_back(avk::context().create_image_view(
        avk::context().create_image(screenRes.x, screenRes.y,
            vk::Format::eD32Sfloat, 1, avk::memory_usage::device,
            avk::image_usage::general_depth_stencil_attachment)));
    result.mShadowMapSamplers.push_back(avk::context().create_image_sampler(
        result.mShadowMaps.back(), sampler));
    attachments.push_back(
        avk::attachment::declare_for(result.mShadowMaps.back().as_reference(),
            avk::on_load::clear.from_previous_layout(avk::layout::undefined),
            avk::usage::depth_stencil, avk::on_store::store));
    result.mShadowMapFramebuffers.push_back(avk::context().create_framebuffer(
        { attachments.back() }, { result.mShadowMaps.back() }));
  }

  assert(nLights > 0);
  result.mShadowPipeline = avk::context().create_graphics_pipeline_for(
      avk::vertex_shader("shaders/shadow.vert"),
      avk::fragment_shader("shaders/shadow.frag"),
      avk::from_buffer_binding(0)->stream_per_vertex<glm::vec3>()->to_location(
          0),
      avk::cfg::front_face::define_front_faces_to_be_counter_clockwise(),
      avk::cfg::viewport_depth_scissors_config::from_framebuffer(
          result.mShadowMapFramebuffers.front().as_reference()),
      attachments.front(),
      avk::push_constant_binding_data{
          avk::shader_type::vertex, 0, sizeof(glm::mat4) });

  result.mLightVPsBuffer
      = avk::context().create_buffer(avk::memory_usage::device, {},
          avk::uniform_buffer_meta::create_from_data(result.mLightVPs));
  auto fillBufferCmd
      = result.mLightVPsBuffer->fill(result.mLightVPs.data(), 0);
  auto fillBufferFence = avk::context().record_and_submit_with_fence(
      { std::move(fillBufferCmd) }, *queue);
  fillBufferFence->wait_until_signalled();

  return result;
}

void Shadows::render(Scene const &scene, avk::queue *queue)
{
  auto const nLights = mLightVPs.size();
  auto &commandPool
      = avk::context().get_command_pool_for_single_use_command_buffers(*queue);
  auto cmdBfrs = commandPool->alloc_command_buffers(
      nLights, vk::CommandBufferUsageFlagBits::eOneTimeSubmit);

  std::vector<avk::fence> fences;
  for (size_t i = 0; i < nLights; ++i)
  {
    fences.push_back(avk::context().create_fence());
    avk::context()
        .record({ avk::command::render_pass(
            mShadowPipeline->renderpass_reference(),
            mShadowMapFramebuffers[i].as_reference(),
            {
                avk::command::bind_pipeline(mShadowPipeline.as_reference()),
                avk::command::custom_commands(
                    [&, this](avk::command_buffer_t &cb)
                    {
                      for (auto &drawCall : scene.mDrawCalls)
                      {
                        cb.record({ avk::command::push_constants(
                                        mShadowPipeline->layout(),
                                        mLightVPs[i]),

                            avk::command::draw_indexed(
                                drawCall.mIndexBuffer.as_reference(),
                                drawCall.mPositionsBuffer.as_reference()) });
                      }
                    }),
            }) })
        .into_command_buffer(cmdBfrs[i])
        .then_submit_to(*queue)
        .signaling_upon_completion(fences[i])
        .submit();
  }
  for (auto &fence : fences)
    fence->wait_until_signalled();
}